### ----------------------------------------------------------------------
#
# The user visible interfaces to drag-and-drop
#

# Example (very stupid) drag handler
proc accepts_drop args {return 1}

# Example generator for strings (for use see demo)
proc supplies_string {w s} {
    global dragdrop

    set dragdrop(supplies,$w) $s
    return STRING
}
# Example dropped data server
proc supply_string {w t} {
    global dragdrop

    if {$t != "STRING"} {
	return -code error "Wrong type code: $t instead of STRING"
    }
    return $dragdrop(supplies,$w)
}

# Sets up widget as a drag-n-drop source
#
# Args are:
#   widget    - What widget should get these abilities?
#               NB. Will probably conflict with other bindings
#   server    - Server to call to get dropped data. Takes
#               {widget name} and {data type} as args
#   generator - Command to run at start of drag operation.
#               Has value of types of data supplied
#   button    - Button to bind
#   notifier  - Server to inform app of current drag-n-drop state.
proc dragdrop_bind {widget server generator {button 3} {notifier {}}} {
    bind $widget <$button> "set dragdropgenerator(%W) \[$generator\]"
    bindtags $widget [concat [bindtags $widget] dragdrop$button]
    dragdrop source $widget datacmd STRING $server
    if [llength $notifier] {
	dragdrop source $widget notify $notifier
    }
}

# Sets up widget as a drag-n-drop destination
#
# Args are:
#   widget           - What widget should get these abilites?
#   dragging_handler - Server to call to find if the widget wants to accept
#                      data at that position, and is also used to implement
#                      look-ahead positioning. Called with all args empty
#                      when the mouse leaves the widget.
#   type             - Type of drop data accepted
#   dropping_handler - Server to call to handle actual drop of data.
proc dragdrop_destination {widget dragging_handler type dropping_handler} {
    dragdrop destination $widget dragcmd $dragging_handler
    dragdrop destination $widget dropcmd $type $dropping_handler
}
