#include "dragdropInt.h"
#include "dragdrop_tab.h"

static int debug = FALSE;
static int dragging = FALSE;
static int Dragdrop_Drag(ClientData, Tcl_Interp *, int, char *[]);

#define DATA ((Dragdrop_data *)data)

static int /* Eval improved (a bit like `eval [concat lp [list ...]]') */
ElEval(Tcl_Interp *interp, char *lead_portion, ...) {
  va_list ap;
  char *s;
  Tcl_DString command;
  int code;

  Tcl_DStringInit(&command);
  Tcl_DStringAppend(&command, lead_portion, ALL);
  va_start(ap, lead_portion);
  while ((s = va_arg(ap, char *)) != EOS) {
    Tcl_DStringAppendElement(&command,s);
  }
  va_end(ap);
  code = Tcl_Eval(interp, V(&command));
  Tcl_DStringFree(&command);
  return code;
}

static int /* Send without version complaints... */
sendcmd(Tcl_Interp *interp, char *app,
	const int argc, char *argv[]) {
  Tcl_DString c;
  int i;

  if (argc<1) {
    interp->result = "tried to send an empty command...";
    return TCL_ERROR;
  }

  Tcl_DStringInit(&c);
  Tcl_DStringAppendElement(&c, "send");
  Tcl_DStringAppendElement(&c, "--");
  Tcl_DStringAppendElement(&c, app);
  Tcl_DStringStartSublist(&c);
  for (i=0;i<argc;i++) Tcl_DStringAppendElement(&c, argv[i]);
  Tcl_DStringEndSublist(&c);

  if ((i=Tcl_Eval(interp, V(&c)))==TCL_ERROR &&
      !strncmp("Tk4", interp->result, 3)) {
    Tcl_ResetResult(interp);
    Tcl_DStringFree(&c);
    return TCL_OK;
  } else {
    Tcl_DStringFree(&c);
    return i;
  }
}

extern int TkGetInterpNames(Tcl_Interp *, Tk_Window);

static int /* finds which widget in which interpreter is the drop-zone */
find_target(Tcl_Interp *interp, char *x, char *y,
	    Tcl_DString *app, Tcl_DString *widget, char *maincmd) {
  /* app and widget must be empty DStrings */
  Tcl_DString s,t;
  int argc,i;
  char **argv, *a[4];

  a[0] = maincmd; a[1] = "containing"; a[2] = x; a[3] = y;

  if (GETVAL(rmt_interp)[0]) { /* Check cache first */
    if (sendcmd(interp,GETVAL(rmt_interp), 4, a)==TCL_OK) {
      Tcl_DStringGetResult(interp, widget);
      Tcl_DStringAppend(app, GETVAL(rmt_interp), ALL);
      return TCL_OK;
    }
    Tcl_ResetResult(interp);
  }
  SETVAL(rmt_interp,""); /* Cache failure, so invalidate */

  /* This next line doesn't exist in the public interface, but it's
   * stupid to cut your nose off to spite your face...
   *
   * In short, this lets me short-cut a call to the interpreter,
   * which would merely end up being translated into a call to this
   * function anyway! POTENTIAL NON-PORTABILITY
   */
  Tcl_DStringInit(&s);
  if (TkGetInterpNames(interp, Tk_MainWindow(interp))==TCL_ERROR) {
    return TCL_ERROR;
  }
  Tcl_DStringGetResult(interp, &s);
  if (Tcl_SplitList(interp, V(&s), &argc, &argv)) {
    Tcl_DStringFree(&s);
    return TCL_ERROR;
  }
  Tcl_DStringFree(&s);
  for (i=0;i<argc;i++) {
    if (debug)
      printf("Trying %s...\n", argv[i]);
    if (sendcmd(interp, argv[i], 4, a)==TCL_OK && interp->result[0]) {
      char *get_patch[2] = { "set", DPATCH };

      Tcl_DStringInit(&t);
      Tcl_DStringGetResult(interp, &t);
      if (sendcmd(interp, argv[i], 2, get_patch)==TCL_OK &&
	  !strcmp(interp->result, VERSION PATCHLEVEL)) {
	Tcl_DStringAppend(app, argv[i], ALL);
	Tcl_DStringAppend(widget, V(&t), ALL);
	free(argv);
	Tcl_DStringFree(&t);
	return TCL_OK;
      }
      Tcl_DStringFree(&t);
    }
    Tcl_ResetResult(interp);
  }

  free(argv);
  /* No application or widget */
  return TCL_OK;
}

Tk_Window
toplevel(const Tk_Window tkwin) {
  Tk_Window w = tkwin;
  while (w) {
    if (Tk_IsTopLevel(w)) {
      return w;
    }
    w = Tk_Parent(w);
  }
  return NULL; /* Really shouldn't get down to here... */
}

/*** ----------------------------------------------------------------------
 * Server procedures
 */

static int /* Start a drag operation */
Dragdrop_Start(ClientData data, Tcl_Interp *interp,
	       int argc, char *argv[]) {
  char *cmd;
  char *arg[4];
  int x,y;
  Tcl_DString name;

  if (argc!=5) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		     " start absoluteX absoluteY typesServed\"", EOS);
    return TCL_ERROR;
  }

  if (Tk_GetPixels(interp, DATA->main, argv[2], &x) != TCL_OK || 
      Tk_GetPixels(interp, DATA->main, argv[3], &y) != TCL_OK) {
    return TCL_ERROR;
  }

  Tcl_DStringInit(&name);
  Tcl_DStringAppend(&name, Tk_PathName(Tk_CoordsToWindow(x,y,DATA->main)),
		    ALL);
  SETVAL(widget, V(&name));
  if (debug)
    printf("%s%s started drag at (%d,%d)\n",
	   Tk_Name(DATA->main), V(&name), x, y);

  SETVAL(types, argv[4]);
  ElEval(interp, V(&name), "cget", "-cursor", EOS);
  SETVAL(cursor, interp->result);
  cmd = DragdropNotify(DATA, V(&name));
  if (!NMT(cmd)) cmd =
    DragdropNotify(DATA,
		   Tk_PathName(toplevel(Tk_NameToWindow(interp,
							V(&name),
							DATA->main))));
  dragging = TRUE; /* switch on the drag and drop commands */
  if (NMT(cmd)) {
    Tcl_ResetResult(interp);
    ElEval(interp, cmd, "startat", V(&name), argv[2], argv[3], EOS);
    if (debug && interp->result[0])
      printf("notify returned >>%s<<\n", interp->result);
  }
  Tcl_DStringFree(&name);

  arg[0] = argv[0];
  arg[1] = "drag";
  arg[2] = argv[2];
  arg[3] = argv[3];
  Tcl_ResetResult(interp);
  return Dragdrop_Drag(data, interp, 4, arg);
}

static int /* Action when dragging (server-side) */
Dragdrop_Drag(ClientData data, Tcl_Interp *interp,
	       int argc, char *argv[]) {
  Tcl_DString a,w;
  char *mywidget, *notify, *Set_Drop[5];
  int ready_for_drop = FALSE;

  if (argc!=4) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		     " drag absoluteX absoluteY\"", EOS);
    return TCL_ERROR;
  }
  if (!dragging)
    return TCL_OK;
  Tcl_DStringInit(&a);
  Tcl_DStringInit(&w);
  if (find_target(interp, argv[2], argv[3], &a, &w, argv[0])==TCL_ERROR) {
    Tcl_DStringFree(&a);
    Tcl_DStringFree(&w);
    return TCL_ERROR;
  }
  mywidget = GETVAL(widget);
  if (!mywidget) {
    Tcl_DStringFree(&a);
    Tcl_DStringFree(&w);
    interp->result = "could not read \""MAIN"(widget)\"";
    return TCL_ERROR;
  }
  notify = DragdropNotify(DATA, mywidget);
  if (!NMT(notify)) notify =
    DragdropNotify(DATA,
		   Tk_PathName(toplevel(Tk_NameToWindow(interp,
							mywidget,
							DATA->main))));

  if (Tcl_DStringLength(&a) && Tcl_DStringLength(&w)) {
    /* Send a clear message if the destination has changed */
    if (strcmp(V(&w), GETVAL(rmt_widget)) ||
	strcmp(V(&a), GETVAL(rmt_interp))) {
      char *cdd[5]; /* Clear drag from remote widget */
      cdd[0] = argv[0];
      cdd[1] = "move";
      cdd[2] = "";
      cdd[3] = "";
      cdd[4] = "";
      sendcmd(interp, V(&a), 5, cdd);
      Tcl_ResetResult(interp);
    }
    /* Send a set message */
    Set_Drop[0] = argv[0];
    Set_Drop[1] = "move";
    Set_Drop[2] = argv[2];
    Set_Drop[3] = argv[3];
    Set_Drop[4] = Tk_Name(DATA->main);
    if (sendcmd(interp, V(&a), 5, Set_Drop)==TCL_OK) {
      Tcl_GetBoolean(interp, interp->result, &ready_for_drop);
    }
    Tcl_ResetResult(interp);
    if (ready_for_drop) {
      SETCURSOR(mywidget,yes);
      SETVAL(rmt_widget, V(&w));
      SETVAL(rmt_interp, V(&a));
      if (NMT(notify)) {
	Tcl_ResetResult(interp);
	ElEval(interp, notify, "able", mywidget, argv[2], argv[3],
	       V(&a), V(&w), EOS);
      }
    } else {
      SETCURSOR(mywidget,no);
      SETVAL(rmt_widget, "");
      SETVAL(rmt_interp, "");
      if (NMT(notify)) {
	Tcl_ResetResult(interp);
	ElEval(interp, notify, "unable", mywidget, argv[2], argv[3], EOS);
      }
    }
  } else {
    SETCURSOR(mywidget,no);
    if (GETVAL(rmt_interp)[0]) {
      char *cdd[5]; /* Clear drag from remote widget */
      cdd[0] = argv[0];
      cdd[1] = "move";
      cdd[2] = "";
      cdd[3] = "";
      cdd[4] = "";
      sendcmd(interp, V(&a), 5, cdd);
    }
    SETVAL(rmt_interp, "");
    SETVAL(rmt_widget, "");
    if (NMT(notify)) {
      Tcl_ResetResult(interp);
      ElEval(interp, notify, "unable", mywidget, argv[2], argv[3], EOS);
    }
  }
  if (debug && interp->result[0])
    printf("notify returned >>%s<<\n", interp->result);
  Tcl_DStringFree(&a);
  Tcl_DStringFree(&w);
  Tcl_ResetResult(interp);
  return TCL_OK;
}

static int /* Action when dropping (server-side) */
Dragdrop_Drop(ClientData data, Tcl_Interp *interp,
	       int argc, char *argv[]) {
  Tcl_DString a,w;
  char *mywidget, *notify, *Set_Drop[5];
  int ready_for_drop = FALSE;
  Tcl_DString result;

  if (argc!=4) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		     " drop absoluteX absoluteY\"", EOS);
    return TCL_ERROR;
  }
  if (!dragging)
    return TCL_OK;
  dragging = FALSE; /* switch off the drag and drop commands */
  Tcl_DStringInit(&a); Tcl_DStringInit(&w);
  if (find_target(interp, argv[2], argv[3], &a, &w, argv[0])==TCL_ERROR) {
    return TCL_ERROR;
  }
  mywidget = GETVAL(widget);
  notify = DragdropNotify(DATA, mywidget);
  if (!NMT(notify)) notify =
    DragdropNotify(DATA,
		   Tk_PathName(toplevel(Tk_NameToWindow(interp,
							mywidget,
							DATA->main))));

  if (Tcl_DStringLength(&a) && Tcl_DStringLength(&w)) {
    Set_Drop[0] = argv[0];
    Set_Drop[1] = "move";
    Set_Drop[2] = argv[2];
    Set_Drop[3] = argv[3];
    Set_Drop[4] = Tk_Name(DATA->main);
    if (sendcmd(interp, V(&a), 5, Set_Drop)==TCL_OK) {
      Tcl_GetBoolean(interp, interp->result, &ready_for_drop);
    }
    { char *cdd[5]; /* Clear drag from remote widget */
      cdd[0] = argv[0];
      cdd[1] = "move";
      cdd[2] = "";
      cdd[3] = "";
      cdd[4] = "";
      sendcmd(interp, V(&a), 5, cdd);
    }
    if (ready_for_drop) {
      Tcl_DStringInit(&result);
      Set_Drop[1] = "set";
      sendcmd(interp, V(&a), 5, Set_Drop);
      Tcl_DStringGetResult(interp, &result);
      if (debug)
	printf("drop succeeded (sent to %s, got >>%s<<)\n", V(&a),
	       V(&result));
      SETCURSOR(mywidget,cursor);
      if (NMT(notify)) {
	Tcl_ResetResult(interp);
	ElEval(interp, notify, "dropped", mywidget,
	       argv[2], argv[3], V(&a), V(&w), EOS);
	if (debug && interp->result[0])
	  printf("notify returned >>%s<<\n", interp->result);
      }
      Tcl_DStringFree(&a);
      Tcl_DStringFree(&w);
      Tcl_ResetResult(interp);
      Tcl_DStringResult(interp, &result);
      return TCL_OK;
    }
  }
  Tcl_DStringFree(&a);
  Tcl_DStringFree(&w);
  SETCURSOR(mywidget,cursor);
  if (debug)
    printf("drop failed. Not over a Tk app?\n");
  SETVAL(types,"");
  if (NMT(notify)) {
    Tcl_ResetResult(interp);
    ElEval(interp, notify, "failed", mywidget, argv[2], argv[3], EOS);
    if (debug && interp->result[0])
      printf("notify returned >>%s<<\n", interp->result);
  }
  Tcl_ResetResult(interp);
  return TCL_OK;
}

static int /* Gets the data being transferred (server-side) */
Dragdrop_GetData(ClientData data, Tcl_Interp *interp,
	       int argc, char *argv[]) {
  char *mywidget, *server;
  Tcl_DString retval;

  if (argc!=3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		     " data type\"", EOS);
    return TCL_ERROR;
  }

  mywidget = GETVAL(widget);
  Tcl_DStringInit(&retval);
  server = DragdropServer(DATA, mywidget, argv[2]);
  if (!NMT(server))
    server =
      DragdropServer(DATA,
		     Tk_PathName(toplevel(Tk_NameToWindow(interp,
							  mywidget,
							  DATA->main))),
		     argv[2]);
  if (NMT(server)) {
    if (!ElEval(interp, server, mywidget, argv[2], EOS)) {
      Tcl_DStringGetResult(interp, &retval);
    }
    if (debug)
      printf("Serving DATA >>%s<< for %s request to %s%s\n",
	     V(&retval), argv[2],
	     Tk_Name(DATA->main), mywidget);
    Tcl_DStringResult(interp, &retval);
  } else {
    interp->result = "";
  }
  return TCL_OK;
}

/*** ----------------------------------------------------------------------
 * Client procedures
 */

static int /* Is the current posn a drop target? (client-side) */
Dragdrop_DragAt(ClientData data, Tcl_Interp *interp,
		int argc, char *argv[]) {
  char xs[NUMSTRSIZE], ys[NUMSTRSIZE], *handler, *tlhandler;
  char *mywidget = GETVAL(rmt_widget);
  Tk_Window thiswin;
  int x,y;

  if (argc!=5) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		     " move absoluteX absoluteY sourceInterp\"", EOS);
    return TCL_ERROR;
  }

  if (argv[2][0] && argv[3][0]) { /* reverse order */
    if (Tk_GetPixels(interp, DATA->main, argv[2], &x) != TCL_OK || 
	Tk_GetPixels(interp, DATA->main, argv[3], &y) != TCL_OK) {
      return TCL_ERROR;
    }

    thiswin = Tk_CoordsToWindow(x, y, DATA->main);
    if (thiswin == NULL)
      goto cleardrag;
    SETVAL(rmt_widget, Tk_PathName(thiswin));
    if (debug)
      printf("%s%s drag at (%s,%s)\n", Tk_Name(DATA->main),
	     Tk_PathName(thiswin), argv[2], argv[3]);
    /*handler = gethandler(interp, Tk_PathName(thiswin));*/
    handler = DragdropDragHandler(DATA, Tk_PathName(thiswin));
    tlhandler = DragdropDragHandler(DATA, Tk_PathName(toplevel(thiswin)));
    if (NMT(handler) || NMT(tlhandler)) {
      int rx,ry;
      if (strcmp(GETVAL(source),argv[4])) {
	char *arg[2];
	arg[0] = "set";
	arg[1] = MAIN"(types)";
	if (sendcmd(interp, argv[4], 2, arg)) return TCL_ERROR;
	SETVAL(types,interp->result);
	SETVAL(source,argv[4]);
      }
      Tk_GetRootCoords(thiswin, &rx, &ry);
      if (Tcl_GetInt(interp, argv[2], &x)) return TCL_ERROR;
      sprintf(xs, "%d", x-rx);
      if (Tcl_GetInt(interp, argv[3], &y)) return TCL_ERROR;
      sprintf(ys, "%d", y-ry);
      Tcl_ResetResult(interp);
      if (NMT(handler)) {
	if (ElEval(interp, handler, Tk_PathName(thiswin), xs, ys,
		   argv[4], GETVAL(types), EOS)) {
	  if (debug)
	    printf("%s\n", interp->result);
	  Tcl_ResetResult(interp);
	  interp->result = "0";
	}
      }
      if (!NMT(interp->result) && NMT(tlhandler)) {
	if (ElEval(interp, tlhandler, Tk_PathName(thiswin), xs, ys,
		   argv[4], GETVAL(types), EOS)) {
	  if (debug)
	    printf("%s\n", interp->result);
	  Tcl_ResetResult(interp);
	  interp->result = "0";
	}
      }	  
    } else {
      Tcl_ResetResult(interp);
      interp->result = "0";
    }
  } else {
  cleardrag:
    if (debug)
      printf("%s%s clear drag\n", Tk_Name(DATA->main), mywidget);
    /*handler = gethandler(interp, mywidget);*/
    handler = DragdropDragHandler(DATA, mywidget);
    { Tk_Window w = Tk_NameToWindow(interp, mywidget, DATA->main);
      tlhandler = w
	? DragdropDragHandler(DATA, Tk_PathName(toplevel(w)))
	: EOS;
    }
    if (NMT(handler))
      ElEval(interp, handler, mywidget, "", "",
	     GETVAL(source), GETVAL(types), EOS);
    else if (NMT(tlhandler))
      ElEval(interp, tlhandler, mywidget, "", "",
	     GETVAL(source), GETVAL(types), EOS);
    SETVAL(source,"");
  }
  return TCL_OK;
}

static int /* Drop data at current posn (client-side) */
Dragdrop_DropAt(ClientData data, Tcl_Interp *interp,
		int argc, char *argv[]) {
  Tk_Window thiswin;
  Tcl_DString ddata,atype;
  int x,y, rx,ry, source = 0;
  char *arg[3], xs[NUMSTRSIZE], ys[NUMSTRSIZE];
  
  if (argc!=5) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		     " set absoluteX absoluteY sourceInterp\"", EOS);
    return TCL_ERROR;
  }
  if (!argv[2][0] || !argv[3][0] || !argv[4][0]) {
    Tcl_AppendResult(interp, "must be nonempty arguments to \"", argv[0],
		     " set\"", EOS);
    return TCL_ERROR;
  }

  if (Tk_GetPixels(interp, DATA->main, argv[2], &x) != TCL_OK || 
      Tk_GetPixels(interp, DATA->main, argv[3], &y) != TCL_OK) {
    return TCL_ERROR;
  }

  thiswin = Tk_CoordsToWindow(x, y, DATA->main);
  if (thiswin == NULL) {
    Tcl_AppendResult(interp, "coordinates not in any window", EOS);
    return TCL_ERROR;
  }

  if (debug)
    printf("%s%s received drop at (%s,%s)\n", Tk_Name(DATA->main),
	   Tk_PathName(thiswin), argv[2], argv[3]);

  /* Get types if necessary */
  if (strcmp(GETVAL(source),argv[4])) {
    char *arg[2];
    arg[0] = "set";
    arg[1] = MAIN"(types)";
    if (sendcmd(interp, argv[4], 2, arg)) return TCL_ERROR;
    SETVAL(types,interp->result);
    SETVAL(source,argv[4]);
  }

  /* Get the matching type */
  Tcl_DStringInit(&atype);
  if (DragdropMatch(interp, DATA, Tk_PathName(thiswin), GETVAL(types))) {
    Tcl_DStringFree(&atype);
    return TCL_ERROR;
  }
  Tcl_DStringGetResult(interp, &atype);

  Tcl_DStringInit(&ddata);
  if (!NMT(V(&atype))) {
    source = 1;
    if (debug)
      printf("trying to find handlers in toplevel for %s\n", GETVAL(types));
    if (DragdropMatch(interp, DATA, Tk_PathName(toplevel(thiswin)),
		      GETVAL(types))) {
      Tcl_DStringFree(&atype);
      return TCL_ERROR;
    }
    Tcl_DStringGetResult(interp, &atype);
  }
  if (!NMT(V(&atype))) {
    if (debug)
      printf("no handlers for %s\n", GETVAL(types));
  } else {
    arg[0] = argv[0];
    arg[1] = "data";
    arg[2] = V(&atype);
    if (sendcmd(interp, argv[4], 3, arg)) {
      Tcl_DStringFree(&atype);
      return TCL_ERROR;
    }
    Tcl_DStringGetResult(interp, &ddata);

    Tk_GetRootCoords(thiswin, &rx, &ry);
    if (Tcl_GetInt(interp, argv[2], &x)) {
      Tcl_DStringFree(&ddata);
      Tcl_DStringFree(&atype);
      Tcl_ResetResult(interp);
      return TCL_OK;
    }
    sprintf(xs, "%d", x-rx);
    if (Tcl_GetInt(interp, argv[3], &y)) {
      Tcl_DStringFree(&ddata);
      Tcl_DStringFree(&atype);
      Tcl_ResetResult(interp);
      return TCL_OK;
    }
    sprintf(ys, "%d", y-ry);

    if (ElEval(interp,
	       DragdropDropHandler(DATA, Tk_PathName(source
						     ? toplevel(thiswin)
						     : thiswin), V(&atype)),
	       Tk_PathName(thiswin), xs, ys, GETVAL(source), V(&atype),
	       V(&ddata), EOS) != TCL_OK) {
      if (debug)
	printf(">>%s<<\n", interp->result);
    }
  }
  Tcl_DStringFree(&ddata);
  SETVAL(source,"");
  SETVAL(types,"");
  Tcl_DStringResult(interp, &atype);
  return TCL_OK;
}

/*** ----------------------------------------------------------------------
 * Configuration and setup
 */
static int
Dragdrop_Cmd(ClientData data, Tcl_Interp *interp,
	     int argc, char *argv[]) {
  char c;
  size_t length;
  Tk_Window window;
  int rootX,rootY, rx,ry, code;

  if (argc < 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"",
		     argv[0], " option ?arg ...?\"", EOS);
    return TCL_ERROR;
  }
  c = argv[1][0];
  length = strlen(argv[1]);
#define SEQ(s,l) ((c==s[0])&&(strncmp(argv[1],s,length)==0)&&(length>l))

  if (SEQ("containing",0)) {
    /* Does a much better job than `winfo containing' by checking that
     * the window that the Tk command would return actually contains
     * the points given, and that the window is in fact mapped!
     */
    if (argc !=4) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		       " containing rootX rootY\"", EOS);
      return TCL_ERROR;
    }
    if ((Tk_GetPixels(interp, DATA->main, argv[2], &rootX) != TCL_OK) ||
	(Tk_GetPixels(interp, DATA->main, argv[3], &rootY) != TCL_OK)) {
      return TCL_ERROR;
    }
    window = Tk_CoordsToWindow(rootX, rootY, DATA->main);
    if (window != NULL) {
      Tk_GetRootCoords(window, &rx, &ry);
      if (Tk_IsMapped(window) && (rootX >= rx) && (rootY >= ry) &&
	  (rootX < rx+Tk_Width(window)) && (rootY < ry+Tk_Height(window))) {
	interp->result = Tk_PathName(window);
      }
    }
  } else if (SEQ("destination",2)){
    /* dragdrop destination widget dragcmd ?cmd?
     * dragdrop destination widget forget
     * dragdrop destination widget dropcmd type ?cmd? ?aftertype?
     * dragdrop destination widget types
     * dragdrop destination widget ?isdestination?
     */
    if (argc<3) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		       " destination widget ?args?\"", EOS);
      return TCL_ERROR;
    }
    if (Tk_NameToWindow(interp, argv[2], DATA->main) == NULL) {
        return TCL_ERROR;
    }
    if (argc==3 ||
	(argc==4 && !strncmp(argv[3], "isdestination", strlen(argv[3])))) {
      Tcl_HashEntry *h = Tcl_FindHashEntry(&(DATA->client), argv[2]);
      interp->result = h ? "1" : "0";
    } else {
      int l = strlen(argv[3]);

      if (argv[3][0] == 'f' && !strncmp(argv[3], "forget", l)) {
	DragdropClntDelete(DATA, argv[2]);
      } else if (argv[3][0] == 't' && !strncmp(argv[3], "types", l)) {
	return DragdropTypes(interp, DATA, argv[2]);
      } else if (argv[3][0] == 'd') {
	if (!strncmp(argv[3], "dragcmd", l) && l>2) {
	  if (argc>5) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
			     " destination ", argv[2], " dragcmd ?command?\"",
			     EOS);
	    return TCL_ERROR;
	  }
	  if (argc==5) {
	    DragdropSetDragHandler(DATA, argv[2], argv[4]);
	  } else {
	    Tcl_AppendResult(interp, DragdropDragHandler(DATA, argv[2]), EOS);
	  }
	} else if (!strncmp(argv[3], "dropcmd", l) && l>2) {
	  char *posn = HANDLER_POS_DEFAULT;

	  switch (argc) {
	  case 5:
	    Tcl_AppendResult(interp,
			     DragdropDropHandler(DATA, argv[2], argv[4]),
			     EOS);
	    break;
	  case 6:
	    if (strlen(argv[5])) {
	      DragdropSetDropHandler(DATA, argv[2], argv[4], argv[5], posn);
	    } else {
	      DragdropSetDropHandler(DATA, argv[2], argv[4], EOS, posn);
	    }
	    break;
	  case 7:
	    if (!strcmp(argv[6],"last")) {
	      posn = HANDLER_POS_END;
	    } else if (!strcmp(argv[6],"first")) {
	      posn = HANDLER_POS_START;
	    } else {
	      posn = argv[6];
	    }
	    if (strlen(argv[5])) {
	      DragdropSetDropHandler(DATA, argv[2], argv[4], argv[5], posn);
	    } else {
	      DragdropSetDropHandler(DATA, argv[2], argv[4], EOS, posn);
	    }
	    break;
	  default:
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
			     " destination ", argv[2], " dropcmd type "
			     "?command? ?aftertype?\"", EOS);
	    return TCL_ERROR;
	  }
	} else goto badDestOpt;
      } else {
      badDestOpt:
	Tcl_AppendResult(interp, "bad option \"", argv[3],
			 "\": must be one of dragcmd, dropcmd, forget, "
			 "isdestination, or types", EOS);
	return TCL_ERROR;
      }
    }
  } else if (SEQ("debug",2)) {
    /* Sets and queries the debugging state
     */
    if (argc > 3) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		       " debug ?boolean?\"", EOS);
      return TCL_ERROR;
    }
    if (argc == 3)
      return Tcl_GetBoolean(interp, argv[2], &debug);
    interp->result = debug ? STRING(1) : STRING(0);
  } else if (SEQ("reset",0)) {
    /* Resets the dragging flag (in case things become confused!)
     */
    if (argc>2) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		       " reset\"", EOS);
      return TCL_ERROR;
    }
    dragging = FALSE;
  } else if (SEQ("source",1)) {
    /* dragdrop source widget notify ?command?
     * dragdrop source widget ?issource?
     * dragdrop source widget forget
     * dragdrop source widget datacmd type ?command?
     * dragdrop source widget types
     */
    if (argc<3) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		       " source widget ?args?\"", EOS);
      return TCL_ERROR;
    }
    if (Tk_NameToWindow(interp, argv[2], DATA->main) == NULL) {
        return TCL_ERROR;
    }
    if (argc==3 ||
	(argc==4 && argv[3][0]=='i' &&
	 !strncmp(argv[3],"issource",strlen(argv[3])))) {
      Tcl_HashEntry *h = Tcl_FindHashEntry(&(DATA->server), argv[2]);
      interp->result = h ? "1" : "0";
    } else {
      int l = strlen(argv[3]);
      if (argv[3][0]=='f' && !strncmp(argv[3],"datacmd",l)) {
	if (argc!=4) {
	  Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
			   " source ", argv[2], " forget\"", EOS);
	  return TCL_ERROR;
	}
	DragdropSrvrDelete(DATA, argv[2]);
      } else if (argv[3][0]=='n' && !strncmp(argv[3], "notify", l)) {
	switch (argc) {
	case 4:
	  Tcl_AppendResult(interp, DragdropNotify(DATA, argv[2]), EOS);
	  break;
	case 5:
	  DragdropSetNotify(DATA, argv[2], argv[4]);
	  break;
	default:
	  Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
			   " source ", argv[2], " notify ?command?\"", EOS);
	  return TCL_ERROR;
	}
      } else if (argv[3][0]=='d' && !strncmp(argv[3], "datacmd", l)) {
	switch (argc) {
	case 5:
	  Tcl_AppendResult(interp,
			   DragdropServer(DATA, argv[2], argv[4]), EOS);
	  break;
	case 6:
	  DragdropSetServer(DATA, argv[2], argv[4], argv[5]);
	  break;
	default:
	  Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
			   " source ", argv[2], " datacmd type ?command?\"",
			   EOS);
	  return TCL_ERROR;
	}
      } else if (argv[3][0]=='t' && !strncmp(argv[3], "types", l)) {
	if (argc!=4) {
	  Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
			   " source ", argv[2], " types\"", EOS);
	  return TCL_ERROR;
	}
	DragdropServerTypes(interp, DATA, argv[2]);
      } else {
	Tcl_AppendResult(interp, "bad option \"", argv[3],
			 "\": must be one of notify, issource, forget, "
			 "datacmd, or types", EOS);
	return TCL_ERROR;
      }
    }
  } else if (SEQ("version",0)) {
    char *s;
    if (argc<2) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		       " version\"", EOS);
      return TCL_ERROR;
    }
    s = Tcl_GetVar(interp, DVERSN, TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
    if (s) {
      interp->result = s;
    } else {
      return TCL_ERROR;
    }
  } else if (SEQ("patchLevel",0)) {
    char *s;
    if (argc<2) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		       " patchLevel\"", EOS);
      return TCL_ERROR;
    }
    s = Tcl_GetVar(interp, DPATCH, TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
    if (s) {
      interp->result = s;
    } else {
      return TCL_ERROR;
    }
  } else if (SEQ("library",0)) {
    char *s;
    if (argc<2) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		       " library\"", EOS);
      return TCL_ERROR;
    }
    s = Tcl_GetVar(interp, DLIBRY, TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
    if (s) {
      interp->result = s;
    } else {
      return TCL_ERROR;
    }
  } else if (SEQ("start",1)) {
    Tk_Preserve(data);
    code = Dragdrop_Start(data, interp, argc, argv);
    Tk_Release(data);
    return code;
  } else if (SEQ("drag",2)) {
    Tk_Preserve(data);
    code =  Dragdrop_Drag(data, interp, argc, argv);
    Tk_Release(data);
    return code;
  } else if (SEQ("drop",2)) {
    Tk_Preserve(data);
    code =  Dragdrop_Drop(data, interp, argc, argv);
    Tk_Release(data);
    return code;
  } else if (SEQ("move",0)) {
    Tk_Preserve(data);
    code =  Dragdrop_DragAt(data, interp, argc, argv);
    Tk_Release(data);
    return code;
  } else if (SEQ("set",0)) {
    Tk_Preserve(data);
    code =  Dragdrop_DropAt(data, interp, argc, argv);
    Tk_Release(data);
    return code;
  } else if (SEQ("data",1)) {
    Tk_Preserve(data);
    code =  Dragdrop_GetData(data, interp, argc, argv);
    Tk_Release(data);
    return code;
  } else {
    Tcl_AppendResult(interp, "bad option \"", argv[1],
		     "\": must be containing, data, debug, destination, "
		     "drag, drop, move, patchLevel, reset, set, source, "
		     "start, or version", EOS);
    return TCL_ERROR;
  }
  return TCL_OK;
}

/* ---------------------------------------------------------------------- */

static void
dddel_callback(ClientData c) {
  DragdropFreeDataTable((Dragdrop_data *)c);
}
static void
dddel(ClientData c) {
  Tk_EventuallyFree(c, dddel_callback);
}

int
INIT(Tcl_Interp *interp) {
  static char initCmd[] =
    "if [file exists [dragdrop library]/init.tcl] {\n\
         source [dragdrop library]/init.tcl\n\
     } else {\n\
         set msg \"can't find [dragdrop library]/init.tcl; perhaps you \"\n\
         append msg \"need to\\ninstall Tcl or set your TCL_LIBRARY \"\n\
         append msg \"environment variable?\"\n\
         error $msg\n\
     }";
  Tcl_DString path;
  ClientData c;
  int i;
  char *lib;

#ifdef MONOCHROME
  if (!GETVAL(yes)) SETVAL(yes, "arrow");
  if (!GETVAL(no))  SETVAL(no,  "circle");
#else
  i = Tk_Visual(Tk_MainWindow(interp)) -> class;
  if (i == StaticGray || i == GrayScale) {
    if (!GETVAL(yes)) SETVAL(yes, "arrow");
    if (!GETVAL(no))  SETVAL(no,  "circle");
  } else {
    if (!GETVAL(yes)) SETVAL(yes, "arrow green1 green4");
    if (!GETVAL(no))  SETVAL(no,  "circle red1 red4");
  }
#endif
  SETVAL(types,"");
  SETVAL(source,"");
  SETVAL(widget,"");
  SETVAL(cursor,"");
  SETVAL(rmt_widget,"");
  SETVAL(rmt_interp,"");

  Tcl_SetVar(interp, DVERSN, VERSION, TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp, DPATCH, VERSION PATCHLEVEL, TCL_GLOBAL_ONLY);

  Tcl_DStringInit(&path);
  Tcl_DStringAppend(&path,
		    Tcl_GetVar(interp, "auto_path", TCL_GLOBAL_ONLY),
		    ALL);
  lib = Tcl_GetVar2(interp, "env", DRAGDROPLIBVAR, TCL_GLOBAL_ONLY);
  if (NMT(lib)) {
    Tcl_DStringAppendElement(&path, lib);
    Tcl_SetVar(interp, DLIBRY, lib, TCL_GLOBAL_ONLY);
  } else {
    Tcl_SetVar(interp, DLIBRY, LIBDIR, TCL_GLOBAL_ONLY);
    Tcl_DStringAppendElement(&path, LIBDIR);
  }
  Tcl_SetVar(interp, "auto_path", V(&path), TCL_GLOBAL_ONLY);
  Tcl_DStringFree(&path);

  c = (ClientData) DragdropAllocDataTable(interp);

  Tcl_CreateCommand(interp, MAINCMD, Dragdrop_Cmd, c, dddel);
  return Tcl_Eval(interp, initCmd);
}
