#include "dragdropInt.h"

typedef Dragdrop_server_table *dst;
typedef Dragdrop_client_table *dct;

#define HANDLER_POS_DEFAULT ((char *)NULL)
#define HANDLER_POS_END     ((char *)1)
#define HANDLER_POS_START   ((char *)2)

static int deleting_everything = 0;

static char *
StrDup(char *s) {
  int i = 1+strlen(s);
  char *d = (char *)ckalloc(sizeof(char)*(i));
  strcpy(d,s);
  return d;
}

/*** ----------------------------------------------------------------------
 * Server-side reading from data tables
 */

#define SV(hep) ((dst)Tcl_GetHashValue(hep))

char *
DragdropNotify(Dragdrop_data *data, char *widget) {
  Tcl_HashEntry *w = Tcl_FindHashEntry(&(data->server), widget);

  if (w == NULL) return NULL;
  return SV(w)->notify;
}

char *
DragdropServer(Dragdrop_data *data, char *widget, char *type) {
  Tcl_HashEntry *w = Tcl_FindHashEntry(&(data->server), widget);
  Tcl_HashEntry *s;

  if (w == NULL) return NULL;
  s = Tcl_FindHashEntry(&(SV(w)->dataserver), type);
  if (s == NULL) return NULL;
  return (char *)Tcl_GetHashValue(s);
}

void
DragdropServerTypes(Tcl_Interp *interp, Dragdrop_data *data, char *widget) {
  Tcl_HashEntry *w = Tcl_FindHashEntry(&(data->server), widget);
  Tcl_HashEntry *s;
  Tcl_HashSearch nxt;

  if (w == NULL) return;
  s = Tcl_FirstHashEntry(&(SV(w)->dataserver), &nxt);
  while (s) {
    Tcl_AppendElement(interp, Tcl_GetHashKey(&(SV(w)->dataserver), s));
    s = Tcl_NextHashEntry(&nxt);
  }
}

/*** ----------------------------------------------------------------------
 * Server-side writing to data tables
 */

void
DragdropSetNotify(Dragdrop_data *data, char *widget, char *notify) {
  int i;
  Tcl_HashEntry *w = Tcl_CreateHashEntry(&(data->server), widget, &i);

  if (i) {
    dst d = (dst)ckalloc(sizeof(Dragdrop_server_table));

    d->notify = StrDup(notify);
    Tcl_InitHashTable(&(d->dataserver), TCL_STRING_KEYS);
    Tcl_SetHashValue(w, (ClientData)d);
  } else {
    if (SV(w)->notify)
      ckfree(SV(w)->notify);
    SV(w)->notify = StrDup(notify);
  }
}

void
DragdropSetServer(Dragdrop_data *data, char *widget,
		   char *type, char *server) {
  int i;
  Tcl_HashEntry *w = Tcl_CreateHashEntry(&(data->server), widget, &i);

  if (i) {
    dst d = (dst)ckalloc(sizeof(Dragdrop_server_table));

    d->notify = (char *)NULL;
    Tcl_InitHashTable(&(d->dataserver), TCL_STRING_KEYS);
    Tcl_SetHashValue(w, (ClientData)d);
    {
      Tcl_HashEntry *t = Tcl_CreateHashEntry(&(d->dataserver), type, &i);

      if (server) {
	char *s = StrDup(server);
	Tcl_SetHashValue(t, (ClientData)s);
      } else {
	Tcl_SetHashValue(t, (ClientData)NULL);
      }
    }
  } else {
    Tcl_HashEntry *t = Tcl_CreateHashEntry(&(SV(w)->dataserver), type, &i);

    if (!i) {
      char *v = (char *)Tcl_GetHashValue(t);
      if (v) ckfree(v);
    }
    if (server) {
      char *s = StrDup(server);
      Tcl_SetHashValue(t, (ClientData)s);
    } else {
      Tcl_SetHashValue(t, (ClientData)NULL);
    }
  }
}

void
DragdropSrvrDelete(Dragdrop_data *data, char *widget) {
  Tcl_HashSearch where;
  Tcl_HashEntry *e = Tcl_FindHashEntry(&(data->server), widget);

  if (e) {
    Tcl_HashEntry *f;
    if (SV(e)->notify) ckfree(SV(e)->notify);
    f = Tcl_FirstHashEntry(&(SV(e)->dataserver), &where);
    while (f != NULL) {
      char *s = (char *)Tcl_GetHashValue(f);
      if (s) ckfree(s);
      f = Tcl_NextHashEntry(&where);
    }
    Tcl_DeleteHashTable(&(SV(e)->dataserver));
    ckfree(SV(e));
    if (!deleting_everything)
      Tcl_DeleteHashEntry(e);
  }
}

/*** ----------------------------------------------------------------------
 * Client-side reading from data tables
 */

#define CV(hep) ((dct)Tcl_GetHashValue(hep))

int
DragdropMatch(Tcl_Interp *interp, Dragdrop_data *data,
	      char *widget, char *typelist) {
  Tcl_HashEntry *w = Tcl_FindHashEntry(&(data->client), widget);
  char **tv; /* Type vector */
  int tc; /* Type count */
  int i,j;

  if (w == NULL)
    return TCL_OK; /* Fail silently */
  if (Tcl_SplitList(interp, typelist, &tc, &tv) != TCL_OK)
    return TCL_ERROR;
  Tcl_ResetResult(interp);
  for (j=0;j<CV(w)->typec;j++) {
    for (i=0;i<tc;i++) {
      if (!strcmp(CV(w)->typev[j],tv[i])) {
	Tcl_HashEntry *h = Tcl_FindHashEntry(&(CV(w)->drophandler), tv[i]);
	if (h == NULL) {
	  Tcl_AppendResult(interp, "interpreter registration mismatch for ",
			   widget, " \"", tv[i], "\"", EOS);
	  ckfree(tv);
	  return TCL_ERROR;
	}
	ckfree(tv);
	Tcl_AppendResult(interp, CV(w)->typev[j], EOS);
	return TCL_OK;
      }
    }
  }
  ckfree(tv);
  return TCL_OK; /* Fail silently */
}

int
DragdropTypes(Tcl_Interp *interp, Dragdrop_data *data, char *widget) {
  Tcl_HashEntry *w = Tcl_FindHashEntry(&(data->client), widget);
  int i;

  Tcl_ResetResult(interp);
  if (w == NULL) return TCL_OK;
  for (i=0;i < CV(w)->typec;i++) {
    Tcl_AppendElement(interp, CV(w)->typev[i]);
  }
  return TCL_OK;
}
char *
DragdropDragHandler(Dragdrop_data *data, char *widget) {
  Tcl_HashEntry *w = Tcl_FindHashEntry(&(data->client), widget);

  if (w == NULL) return NULL;
  return CV(w)->draghandler;
}
char *
DragdropDropHandler(Dragdrop_data *data, char *widget, char *type) {
  Tcl_HashEntry *w = Tcl_FindHashEntry(&(data->client), widget);

  if (w == NULL) return NULL;
  else {
    Tcl_HashEntry *h = Tcl_FindHashEntry(&(CV(w)->drophandler), type);

    if (h == NULL) return NULL;
    return (char *)Tcl_GetHashValue(h);
  }
}

/*** ----------------------------------------------------------------------
 * Client-size writing to data tables
 */

void
DragdropSetDragHandler(Dragdrop_data *data, char *widget, char *handler) {
  int i;
  Tcl_HashEntry *w = Tcl_CreateHashEntry(&(data->client), widget, &i);

  if (i) {
    dct d = (dct)ckalloc(sizeof(Dragdrop_client_table));

    d->draghandler = StrDup(handler);
    Tcl_InitHashTable(&(d->drophandler), TCL_STRING_KEYS);
    d->typec = 0;
    d->typecmax = 0;
    d->typev = NULL;
    Tcl_SetHashValue(w, (ClientData)d);
  } else {
    if (CV(w)->draghandler)
      ckfree(CV(w)->draghandler);
    CV(w)->draghandler = StrDup(handler);
  }
}

#define HANDLER_TAB_INCR    8

void
DragdropSetDropHandler(Dragdrop_data *data, char *widget, char *type,
		       char *handler, char *after) {
  int i;
  Tcl_HashEntry *w = Tcl_CreateHashEntry(&(data->client), widget, &i);

  if (i) {
    dct d = (dct)ckalloc(sizeof(Dragdrop_client_table));

    d->draghandler = EOS;
    Tcl_InitHashTable(&(d->drophandler), TCL_STRING_KEYS);
    d->typec = 1;
    d->typecmax = HANDLER_TAB_INCR;
    d->typev = (char **)ckalloc(sizeof(char *)*HANDLER_TAB_INCR);
    d->typev[0] = StrDup(type);

    Tcl_SetHashValue(w, (ClientData)d);
    {
      Tcl_HashEntry *t = Tcl_CreateHashEntry(&(d->drophandler), type, &i);

      if (handler) {
	char *s = StrDup(handler);
	Tcl_SetHashValue(t, (ClientData)s);
      } else {
	Tcl_SetHashValue(t, (ClientData)NULL);
      }
    }
  } else if (handler == NULL || !handler[0]) {
    Tcl_HashEntry *t = Tcl_FindHashEntry(&(CV(w)->drophandler), type);

    if (t) {
      /* Remove from type list */
      int i,j;
      ckfree(Tcl_GetHashValue(t));
      Tcl_DeleteHashEntry(t);
      for (i=j=0;i<CV(w)->typec;i++) {
	if (j) CV(w)->typev[i-1] = CV(w)->typev[i];
	if (!j && !strcmp(type, CV(w)->typev[i])) j=1;
      }
      CV(w)->typec--;
    }
  } else {
    Tcl_HashEntry *t = Tcl_CreateHashEntry(&(CV(w)->drophandler), type, &i);
    char *s = StrDup(handler);

    if (!i) {
      char *v = (char *)Tcl_GetHashValue(t);
      if (v) ckfree(v);
    }
    Tcl_SetHashValue(t, (ClientData)s);

    if (i) {
      /* Add to type list */
      char *ty = StrDup(type);
      if (CV(w)->typec<CV(w)->typecmax) {
	/* Shuffle */
	if (after == HANDLER_POS_DEFAULT ||
	    after == HANDLER_POS_END) {
	  CV(w)->typev[CV(w)->typec] = ty;
	  CV(w)->typec++;
	} else {
	  int x,f;
	  for (x=CV(w)->typec,f=1; f&&x>0; x--)
	    if (after != HANDLER_POS_START &&
		!strcmp(after, CV(w)->typev[x-1]))
	      f=0;
	    else {
	      CV(w)->typev[x] = CV(w)->typev[x-1];
	    }
	  CV(w)->typev[x] = ty;
	  CV(w)->typec++;
	}
      } else {
	/* Must allocate new */
	int newsize = CV(w)->typecmax+HANDLER_TAB_INCR;
	char **ntypev = (char **)ckalloc(sizeof(char *)*newsize);
	int o = CV(w)->typec;
	int n = CV(w)->typec;

	if (after == HANDLER_POS_DEFAULT ||
	    after == HANDLER_POS_END) {
	  ntypev[n--] = ty;
	}
	while ((o--)>0) {
	  if (after != HANDLER_POS_DEFAULT &&
	      after != HANDLER_POS_END &&
	      after != HANDLER_POS_START &&
	      !strcmp(after, CV(w)->typev[o])) {
	    ntypev[n--] = ty;
	  }
	  ntypev[n--] = CV(w)->typev[o];
	}
	if (after == HANDLER_POS_START) {
	  ntypev[0] = ty;
	}
	ckfree(CV(w)->typev);
	CV(w)->typev = ntypev;
	CV(w)->typec++;
      }
    } else {
      /* Move within type list */
      if (after == HANDLER_POS_END) {
	after = CV(w)->typev[CV(w)->typec-1];
      }
      if (!strcmp(after,type)) {
	char *tmp = EOS, *t2 = EOS;
	int i,j,k;

	if (after == HANDLER_POS_START) {
	  /* Move type to beginning */
	  tmp = CV(w)->typev[0];
	  for (i=1;i<CV(w)->typec;i++,j++) {
	    if (!strcmp(CV(w)->typev[i],type)) {
	      CV(w)->typev[0] = CV(w)->typev[i];
	      CV(w)->typev[i] = tmp;
	      break;
	    } else {
	      t2 = tmp;
	      tmp = CV(w)->typev[i];
	      CV(w)->typev[i] = t2;
	    }
	  }
	} else if (after != HANDLER_POS_DEFAULT) {
	  /* Move type to after `after' */
	  for (i=0;i<CV(w)->typec;i++) {
	    if (!strcmp(CV(w)->typev[i],type)) {
	      /* found type first, must search for after while shuffling */
	      tmp = CV(w)->typev[i];
	      for (j=i+1;j<CV(w)->typec;j++) {
		CV(w)->typev[j-1] = CV(w)->typev[j];
		if (!strcmp(CV(w)->typev[j],after)) {
		  CV(w)->typev[j] = tmp;
		  break;
		}
	      }
	      break;
	    } else if (!strcmp(CV(w)->typev[i],after)) {
	      /* found after first, must search (backwards from end)
               * for type while shuffling.
	       */
	      for (j=CV(w)->typec-1;j>i;j--)
		if (!strcmp(CV(w)->typev[j],type)) {
		  tmp = CV(w)->typev[j];
		  for (k=j-1;k>i;k--)
		    CV(w)->typev[k+1] = CV(w)->typev[k];
		  CV(w)->typev[i+1] = tmp;
		}
	      break;
	    }
	  }
	} 
      }
    }
  }
}

void
DragdropClntDelete(Dragdrop_data *data, char *widget) {
  Tcl_HashSearch where;
  Tcl_HashEntry *e = Tcl_FindHashEntry(&(data->client), widget);

  if (e) {
    Tcl_HashEntry *f;
    if (CV(e)->draghandler) ckfree(CV(e)->draghandler);
    if (CV(e)->typecmax) ckfree(CV(e)->typev);
    f = Tcl_FirstHashEntry(&(CV(e)->drophandler), &where);
    while (f != NULL) {
      char *s = (char *)Tcl_GetHashValue(f);
      if (s) ckfree(s);
      f = Tcl_NextHashEntry(&where);
    }
    Tcl_DeleteHashTable(&(CV(e)->drophandler));
    ckfree(CV(e));
    if (!deleting_everything)
      Tcl_DeleteHashEntry(e);
  }
}

/*** ----------------------------------------------------------------------
 * Dealing with whole data tables...
 */

Dragdrop_data *
DragdropAllocDataTable(Tcl_Interp *interp) {
  Dragdrop_data *d = (Dragdrop_data *)ckalloc(sizeof(Dragdrop_data));
  
  Tcl_InitHashTable(&(d->server), TCL_STRING_KEYS);
  Tcl_InitHashTable(&(d->client), TCL_STRING_KEYS);
  d->main = Tk_MainWindow(interp);
  return d;
}
void
DragdropFreeDataTable(Dragdrop_data *data) {
  Tcl_HashEntry *he;
  Tcl_HashSearch s;

  deleting_everything = 1;
  he = Tcl_FirstHashEntry(&(data->server), &s);
  while (he) {
    DragdropSrvrDelete(data, Tcl_GetHashKey(&(data->server), he));
    he = Tcl_NextHashEntry(&s);
  }
  he = Tcl_FirstHashEntry(&(data->client), &s);
  while (he) {
    DragdropClntDelete(data, Tcl_GetHashKey(&(data->client), he));
    he = Tcl_NextHashEntry(&s);
  }
  deleting_everything = 0;
  Tcl_DeleteHashTable(&(data->server));
  Tcl_DeleteHashTable(&(data->client));
  ckfree(data);
}
